/**
Object corresponding to a row of the constraint table.

@author Michael Breen
*/
class ConstraintRow
{
   /** A single constraint from the constraint table. */
   private static class TableConstraint extends Constraint
   {
      /** The constraining values (shaded table cell). */
      private ValueSet ifVals;

      /** The constrained values. */
      private ValueSet thenVals;

      TableConstraint(ValueSet ifVals, ValueSet thenVals)
      {
         this.ifVals = ifVals;
         this.thenVals = thenVals;
      }

      public boolean allows(Combination comb)
      {
         return comb.getValue(
            this.thenVals.getVariable()).in(this.thenVals) ||
            !comb.getValue(
            this.ifVals.getVariable()).in(this.ifVals);
      }

      /**
      @return a string in the form, e.g.,
      "Constraint Table (X = x1 => Y = y1 | y2)"
      */
      public String toString()
      {
         StringBuffer buf = new StringBuffer(
            "Constraint Table (");
         buf.append(this.ifVals);
         buf.append(" => ");
         buf.append(this.thenVals);
         buf.append(")");
         return buf.toString();
      }
   }

   /** The constraining variable (identifies the "if" values). */
   private Variable ifVar;

   /**
   The combinations allowed by this row are those allowed
   by this mask, together with any combination in which the
   constraining variable does not have one of the values
   in the mask.
   */
   private CombinationMask rowMask = new CombinationMask();

   /**
   @param constrainingVar the constraining variable for
   this row
   */
   ConstraintRow(Variable constrainingVar)
   {
      this.ifVar = constrainingVar;
   }

   /**
   @param comb combination to be checked
   @return a table constraint which renders the argument
   invalid; null if there is no such constraint in this row
   */
   Constraint invalidating(Combination comb)
   {
      ValueSet ifVals = this.rowMask.getValues(this.ifVar);
      if (!comb.getValue(this.ifVar).in(ifVals))
         return null;
      for (int i = 0; i < Variable.numVars(); ++i)
      {
         Variable var = Variable.getVariable(i);
         if (var != this.ifVar)
         {
            ValueSet thenVals = this.rowMask.getValues(var);
            if (!comb.getValue(var).in(thenVals))
               return new TableConstraint(ifVals, thenVals);
         }
      }
      return null;
   }

   /**
   @param values the set of values for a variable - either
   the constraining variable or one of the constrained
   variables
   @return this object (to facilitate invocation chaining)
   */
   ConstraintRow setValues(ValueSet values)
   {
      this.rowMask.setValues(values);
      return this;
   }
}
