import java.util.HashMap;

/**
Provides methods which indicate whether a particular
combination of variable values, or state, is permitted
by the constraints and whether an event is among those
expected to happen in a particular state.

@author Michael Breen
*/
public final class Constraints
{
   private Constraints() { }

   /**
   Check whether a particular combination of variable
   values is allowed by the constraints.
   @param comb combination of variable values (state)
   to be checked
   @return true if the global constraints (including any
   constraints in the constraint table) permit the
   argument, otherwise false
   */
   public static boolean valid(Combination comb)
   {
      return explainInvalid(comb) == null;
   }

   /**
   Checks if the constraints for an event allow it to
   happen in a particular state.
   Note this does not check the global constraints to see
   whether the state itself is valid.
   @param event an event defined in the model
   @param comb a combination of variable values (state)
   @return true if the event is one that is expected
   in the specified state, otherwise false
   */
   public static boolean expected(Event event, Combination comb)
   {
      return explainUnexpected(event, comb) == null;
   }

   /**
   Find out why a particular combination of variable values
   is invalid.
   @param comb combination of variable values (state)
   to be checked
   @return the first constraint found which makes the
   argument invalid; null if there is no such constraint
   */
   public static Constraint explainInvalid(Combination comb)
   {
      for (int i = 0; i < Constraints.table.length; ++i)
      {
         Constraint constraint =
            Constraints.table[i].invalidating(comb);
         if (constraint != null)
            return constraint;
      }
      for (int i = 0; i < Constraints.globals.length; ++i)
      {
         if (!Constraints.globals[i].allows(comb))
            return Constraints.globals[i];
      }
      return null;
   }

   /**
   Find out why an event is not expected in a particular
   state.
   @param event an event defined in the model
   @param comb combination of variable values (state)
   to be checked
   @return the first constraint associated with the
   event which makes it unexpected in the specified state;
   null if there is no such constraint
   */
   public static Constraint explainUnexpected(Event event,
      Combination comb)
   {
      Constraint[] constraints = (Constraint[])
         Constraints.eventConstraints.get(event);
      if (constraints == null)
         return null;
      for (int i = 0; i < constraints.length; ++i)
      {
         if (!constraints[i].allows(comb))
            return constraints[i];
      }
      return null;
   }

   /** Implements a constraint expressed as a formula. */
   private static class DefaultConstraint extends Constraint
   {
      Event event;
      String id;

      /** Constructor for a global constraint. */
      DefaultConstraint(String name)
      {
         this(null, name);
      }

      /** Constructor for an event constraint. */
      DefaultConstraint(Event event, String name)
      {
         this.event = event;
         this.id = name;
      }

      /** This returns true unless overridden. */
      public boolean allows(Combination comb) { return true; }

      /**
      @return this constraint's identifier in the model
      if it is a global constraint or the identifier
      preceded by the event it is an event constraint
      */
      public String toString()
      {
         if (this.event != null)
            return "[" + this.event + "] " + this.id;
         else
            return this.id;
      }
   }

   /** Constraint table. */
   static ConstraintRow[] table = {
<#list doc.model.simple_constraints.constraint_group as row>
   <@constraint_row_init row/>
</#list>
      };

   /** Global constraint formulas. */
   static Constraint[] globals = {
<#list doc.model.constraints.constraint as constraint>
   <@constraint_init constraint/>
</#list>
      };

   /**
   Maps each event to an array of constraints, if it has any.
   */
   static HashMap eventConstraints = new HashMap();
<#list doc.model.events.event as event>
   <#if (event["./constraints/constraint"]?size > 0)>

   /** Constraints for event '${event.@name}'. */
   private static Constraint[] constraints${event_id(event)} = {
      <#list event.constraints.constraint as constraint>
         <@constraint_init constraint/>
      </#list>
      };
   </#if>
</#list>

   // initialise eventConstraints
   static
   {
<#list doc.model.events.event as event>
   <#if (event["./constraints/constraint"]?size > 0)>
      Constraints.eventConstraints.put(Event.${event_id(event)},
         constraints${event_id(event)});
   </#if>
</#list>
   }
}
<#--------------------  Local Macros  -------------------->

<#--
Output code to initialise a row of the constraint table
(no output unless the row contains at least one constraint).
-->
<#macro constraint_row_init row>
   <#if (row[
"./value_set[./@type='constrained'][not(./@content='dont_care')]"]
?size > 0)
   >
      new ConstraintRow(Variable.${variable_id(
row["./value_set[./@type='constraining']/@variable"][0])})<#--
-->
      <#list row["./value_set[not(./@content='dont_care')]"]
       as vals
      >

      .setValues(<@values_to_code vals/>)<#--
-->
      </#list>
,
   </#if>
</#macro>

<#-- output code to create a constraint -->
<#macro constraint_init constraint>
      new DefaultConstraint(<#--
-->
   <#local events = constraint?ancestors("event")/>
   <#if events?size != 0
    >Event.${event_id(events[0])}, </#if
   >"${constraint.@name}")<#--
-->
   <#if constraint.formula.parse_node?size != 0>

      {
         public boolean allows(Combination comb)
         {
      <@parse_node_to_code node=constraint.formula.parse_node />
            return true;
         }
      }<#--
-->
   </#if>
,
</#macro>
