import java.util.ArrayList;
import java.util.HashMap;

/**
The main interface to the logic in the model.
Given an event and a combination of variable values
(in a finite state machine model, this is the current
state), {@link Rules#applyRules Rules.applyRules()}
finds the appropriate rule, if any, and updates
the state.
<p>
<b>Constraints are ignored</b>: it is
not checked that states are valid or that an
event is expected in a particular state.
If you need to do these checks, refer to the
{@link Constraints} class.
<p>
Typically, {@link Rules#applyRules Rules.applyRules()}
will be called as part of a loop within the application,
where the state is repeatedly updated in response to the
latest detected event.
The code in this loop could be something like:
<pre>
   // call local method to get the latest event
   Event event = waitForEvent();
   // copy the current state before it changes
   Combination oldState = new Combination(state);
   Rule rule = Rules.applyRules(event, state, null);
   if (rule != null)
   {
      // Our application is a digital alarm clock;
      // 'Alarms' is a variable defined in the model...
      if (state.get$Alarms() != oldState.get$Alarms())
      {
         if (oldState.get$Alarms() == $Alarms.$none)
         {
            // update the clock's display
            display.setElement(Display.BELL_SYMBOL,
               Display.ENABLED);
         }
         // ...
      }
      // ...
   }
   // ...
</pre>

@author Michael Breen
*/
public class Rules
{
   /**
   Callback interface for code integration.
   Where the applicability or the outcome of a rule
   depends in part on something outside the model,
   for example, the value of a numeric variable,
   this interface can be implemented by external code
   which implements the necessary checks.
   <p>
   Here's an example.
   Suppose you have a model for a digital alarm clock
   which allows a number of different alarm times
   to be set.
   One of the rules might be:
   <blockquote>
   <b>Event Button_D, Rule set_alarm_1b:</b>
   <TABLE border=1 summary="example of a rule with extra logic">
   <THEAD><TR>
   <TH>Mode</TH>
   <TH>Alarms</TH>
   <TH>(Other Vars...)</TH>
   </TR></THEAD>
   <TBODY>
   <TR>
   <TD class="source">program</TD>
   <TD class="source">none<BR>some</TD>
   <TD class="source">...</TD>
   </TR>
   <TR class="target-row">
   <TD class="target">=</TD>
   <TD class="target">some<BR>max</TD>
   <TD class="target">...</TD>
   </TR>
   <TR><TD colspan=4>
   // The selected time is not the time of an<BR>
   // existing alarm; a new alarm is set to go<BR>
   // off at the selected time.
   </TD></TR>
   </TBODY>
   </TABLE>
   </blockquote>
   Here, part of the rule's logic exists only as a comment
   in the model: code to implement this logic must be added
   to the generated code.
   <p>
   Also, the new value of the 'Alarms' variable is uncertain
   in the rule and so the added code must determine whether
   this variable changes to 'some' or, if the maximum number
   of alarms has been set, 'max'.
   <p>
   Code corresponding to the comment would then be
   integrated with the generated code like this:
   <pre>
   Rules.ExtraLogic myLogic = new Rules.ExtraLogic()
      {
         public boolean confirmRule(Rule rule,
            Combination current, Combination next)
         {
            if (rule == Rule.Button_D$set_alarm_1b)
            {
               AlarmList alarms = AlarmList.getAlarms();
               ClockTime selectedTime = Clock.getSelectedTime();
               for (int i = 0; i &lt; alarms.size(); ++i)
               {
                  if (selectedTime.equals(alarms.get(i)))
                  {
                     // this rule does not apply;
                     // go back and look for another rule
                     return false;
                  }
               }
               alarms.add(selectedTime);
               // the next value of 'Alarms' will have been set
               // arbitrarily to either 'some' or 'max' in the
               // generated code; set the correct value here
               if (alarms.size() &lt; AlarmList.MAX_ALARMS)
                  next.setValue($Alarms.$some);
               else
                  next.setValue($Alarms.$max);
               return true;
            }
            // else if (rule == ...)
            // ...
            // else if (rule.getEvent() == ...)
            // ...
            else
            {
               return true;
            }
         }
      };

   Rule rule = {@link Rules#applyRules Rules.applyRules}(event, state, myLogic);
   </pre>
   <p>
   Note: In this example, the code to ensure that the new
   value of the 'Alarms' variable is correct in the
   comfirmRule() method could be replaced by code just after
   the call to Rules.applyRules().
   In other words, after each call to applyRules(),
   the value of the 'Alarms' variable in 'state' would be
   set to 'none', 'some' or 'max' depending on how many
   alarms were in the list of alarms.
   */
   public static interface ExtraLogic
   {
      /**
      {@link Rules#applyRules Rules.applyRules()}
      calls this method to give
      external code a chance to veto a rule.
      For a non-deterministic rule (that is, a rule
      which allows more than one possible next state
      for a given current state), the next state will
      have been chosen arbitrarily and this method
      is free to change it.
      <p>
      <b>Note:</b> If this method changes 'next' then
      it is responsible for ensuring that it remains
      consistent with the logic of the rule
      (this can be double-checked by calling the
      {@link Rule#applies applies()}
      method of the rule with the altered
      next state) and with the constraints.
      <p>
      The event can be derived from the rule.
      @param rule a rule which seems to apply in the
      current state
      @param current the current state
      @param next the next state as determined by the
      rule (can be changed here - see above)
      @return true to accept the rule, false to veto it
      and search for another rule
      */
      public boolean confirmRule(Rule rule,
         Combination current, Combination next);
   }

   /** Maps each event to its list of rules. */
   static final HashMap rules = new HashMap();
   static
   {
      for (int i = 0; i < Event.numEvents(); ++i)
         Rules.rules.put(Event.getEvent(i), new ArrayList());
   }

   /**
   If any rule applies for the specified event and
   current combination of variable values (state) then
   change the state according to the first such rule
   found and return that rule.
   <p>
   Note: If, for the specified current state, the rule
   allows for more than one possible next state
   then the next state is chosen arbitrarily.
   Even if only some of the possible next states are valid,
   no attempt is made to choose a valid next state
   in preference to an invalid one.
   @param event the event which has just occurred
   @param state the current combination of variable values,
   or state; if a rule that applies in this state is found
   then <b>this object is changed</b> to the next state as
   determined by the rule
   @param extraLogic this object is given a chance to veto
   any rule that appears to apply; can be null
   @return the rule that was found or null if no rule applies
   */
   public static Rule applyRules(Event event, Combination state,
      ExtraLogic extraLogic)
   {
      ArrayList list = (ArrayList) Rules.rules.get(event);
      for (int i = 0; i < list.size(); ++i)
      {
         Rule rule = (Rule) list.get(i);
         Combination next = rule.apply(state);
         if (next != null)
         {
            if (extraLogic != null)
            {
               if (!extraLogic.confirmRule(rule, state, next))
                  continue;
            }
            state.setEqualTo(next);
            return rule;
         }
      }
      return null;
   }

   /**
   Only rules registered via this method will be considered
   when trying to determine the response to an event.
   @param rule rule to be considered on receipt of an event
   */
   static void addRule(Rule rule)
   {
      ArrayList list = (ArrayList) Rules.rules.get(
         rule.getEvent());
      list.add(rule);
   }

   /*
   Force initialisation of class Rule so that each of
   the static Rule objects is created and added to the
   local lists of rules.
   */
   static
   {
      try
      {
         Class.forName("${package_prefix()}Rule");
      }
      catch (ClassNotFoundException e)
      {
      }
   }

   private Rules() { }
}
