import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;

/**
Sample code to illustrate some of the API provided by
the generated code.
This assumes the Statestep model works as a finite
state machine. For a decision table, the updated 'state'
would be ignored; instead, the combination passed to
Rules.applyRules() would be newly determined each time.

@author Michael Breen
*/
public class SampleCode
{
   static BufferedReader reader = new BufferedReader(
      new InputStreamReader(System.in));

   /**
   Repeatedly request an event, determine the response
   to the event and echo the updated state.
   */
   public static void main(String args[])
      throws IOException
   {
      if (Variable.numVars() == 0 || Event.numEvents() == 0)
      {
         System.out.println("This demo requires code "
            + "generated from a model with at least one "
            + "variable for which values have been defined "
            + "and at least one event.");
         return;
      }
      Combination state = new Combination();
      // Try to make the initial state a valid combination
      for (long i = 0; !Constraints.valid(state) &&
         i < Combination.numCombinations(); ++i)
      {
         state = Combination.getCombination(i);
      }
      System.out.println("Initial state: " + state);
      for (;;)
      {
         if (!Constraints.valid(state))
         {
            System.out.println();
            System.out.println("This state is invalid! "
               + "It violates constraint \""
               + Constraints.explainInvalid(state) + "\".");
         }
         Event event = readEvent();
         System.out.println();
         System.out.println("Event '" + event + "' occurs...");
         if (!Constraints.expected(event, state))
         {
            System.out.println();
            System.out.println("This event not expected in "
               + "this state! "
               + "Its occurrence violates constraint \""
               + Constraints.explainUnexpected(event, state)
               + "\".");
            System.out.println("Event ignored.");
         }
         else
         {
            Combination oldState = new Combination(state);
            Rule rule = Rules.applyRules(event, state, null);
            if (rule == null)
            {
               System.out.println("No rule applies.");
            }
            else
            {
               System.out.println("Rule '" + rule + "' applies.");
               checkStateChange(oldState, state);
            }
         }
         System.out.println();
         System.out.println("Current state: " + state);
      }
   }

   /**
   Indicate whether or how the state has changed.
   */
   static void checkStateChange(Combination one, Combination two)
   {
      System.out.println();
      if (one.equals(two))
      {
         System.out.println("The state has not changed.");
      }
      else
      {
         System.out.print("The values of the following "
            + "variables changed:");
         for (int i = 0; i < Variable.numVars(); ++i)
         {
            Variable var = Variable.getVariable(i);
            if (one.getValue(var) != two.getValue(var))
               System.out.print(" " + var);
         }
         System.out.println();
      }
   }

   /**
   Read an event from the standard input.
   */
   static Event readEvent()
      throws IOException
   {
      System.out.println();
      for (;;)
      {
         System.out.print("Enter event [0 - "
            + (Event.numEvents() - 1) + "], 'l' to List, "
            + "or 'q' to Quit: ");
         String line = reader.readLine();
         if (line == null)
            System.exit(0);
         line = line.trim().toLowerCase();
         try
         {
            int num = Integer.parseInt(line);
            if (num >= 0 && num < Event.numEvents())
               return Event.getEvent(num);
            System.out.println(num +
               " is not the number of a valid event.");
         }
         catch (NumberFormatException nfe)
         {
            if (line.equals("l"))
            {
               for (int i = 0; i < Event.numEvents(); ++i)
               {
                  System.out.println(i + ": "
                     + Event.getEvent(i));
               }
            }
            else if (line.equals("q"))
            {
               System.exit(0);
            }
            else
            {
               System.out.println("Unexpected input \"" + line
                  + "\".");
            }
         }
      }
   }
}
