import java.util.BitSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
A set of one or more values of the same variable.
Instances of this class are immutable.

@author Michael Breen
*/
public abstract class ValueSet
{
   /** Iterator over the values in this set. */
   private class ValueSetIterator implements Iterator
   {
      /**
      The bit corresponding to the last value returned
      or -1 if no value has yet been returned.
      */
      private int index = -1;

      public ValueSetIterator() { }

      public boolean hasNext()
      {
         return this.index <
            ValueSet.this.getValueBits().length() - 1;
      }

      public Object next()
      {
         if (!hasNext())
            throw new NoSuchElementException();
         while (!ValueSet.this.getValueBits().get(++this.index));
         return ValueSet.this.getVariable().getValue(this.index);
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }

   /** The standard implementation of a ValueSet. */
   private static class DefaultValueSet extends ValueSet
   {
      private BitSet valueBits;

      private DefaultValueSet(Variable variable, BitSet vals)
      {
         super(variable);
         this.valueBits = vals;
      }

      BitSet getValueBits() { return this.valueBits; }
   }

   /** The variable to which the values in this set belong. */
   private Variable variable;

   /** @param variable the variable to which the values belong */
   ValueSet(Variable variable) { this.variable = variable; }

   /**
   A set of values is implemented as a BitSet in which
   the bits corresponding to its elements are set.
   @return the set of bits corresponding to the set of
   values in this set
   */
   abstract BitSet getValueBits();

   /**
   Get the union of this set with another set.
   @param operand a set of values to OR with this set
   @return the union of the two sets of values
   @throws IllegalArgumentException if the values in
   the two sets belong to different variables
   */
   public ValueSet or(ValueSet operand)
   {
      if (operand.getVariable() != this.variable)
         throw new IllegalArgumentException();
      BitSet bits = (BitSet) getValueBits().clone();
      bits.or(operand.getValueBits());
      return new DefaultValueSet(this.variable, bits);
   }

   /**
   Get the intersection of this set with another set.
   @param operand a set of values to AND with this set
   @return the intersection of the two sets of values
   if that is not the empty set, otherwise null
   @throws IllegalArgumentException if the values in
   the two sets belong to different variables
   */
   public ValueSet and(ValueSet operand)
   {
      if (operand.getVariable() != this.variable)
         throw new IllegalArgumentException();
      BitSet bits = (BitSet) getValueBits().clone();
      bits.and(operand.getValueBits());
      if (bits.length() > 0)
         return new DefaultValueSet(this.variable, bits);
      else
         return null;
   }

   /**
   Get the set of all elements in this set which are
   not also elements in another set.
   @param operand a set of values to exclude
   @return the values in this set which are not in the
   operand or null if there are no such values
   @throws IllegalArgumentException if the values in
   the two sets belong to different variables
   */
   public ValueSet andNot(ValueSet operand)
   {
      if (operand.getVariable() != this.variable)
         throw new IllegalArgumentException();
      BitSet bits = (BitSet) getValueBits().clone();
      bits.andNot(operand.getValueBits());
      if (bits.length() > 0)
         return new DefaultValueSet(this.variable, bits);
      else
         return null;
   }

   /**
   Get the variable of which all the elements in this set
   are values.
   @return the variable to which these values belong
   */
   public Variable getVariable() { return this.variable; }

   /**
   Get an iterator over the values in this set.
   @return an iterator over the values in this set
   */
   public Iterator iterator() { return new ValueSetIterator(); }

   /**
   The string representation of a set of values takes the
   form "varName = valName1 | valName2 | ...".
   @return a string describing this object
   */
   public String toString()
   {
      StringBuffer buf = new StringBuffer(
         getVariable().toString());
      buf.append(" = ");
      Iterator iter = iterator();
      for (;;)
      {
         buf.append(((Value) iter.next()).getName());
         if (iter.hasNext())
            buf.append(" | ");
         else
            break;
      }
      return buf.toString();
   }

   /**
   Two sets are equal if they contain the same elements.
   @param obj the object to be compared with this object
   @return true if the argument is a ValueSet which contains
   the same values as this set, otherwise false
   */
   public boolean equals(Object obj)
   {
      if (!(obj instanceof ValueSet))
         return false;
      ValueSet arg = (ValueSet) obj;
      return this.variable == arg.variable
         && getValueBits().equals(arg.getValueBits());
   }

   public int hashCode() { return getValueBits().hashCode(); }
}
