<#-- Macros and functions to be used in the template files -->

<#--
The following _id functions generate identifiers to use
in the code from the names of the corresponding
entities (e.g., variables, events) in the model.
In general, they are changed slightly in case they
correspond to Java keywords (e.g., a variable might
be called 'Object').
This is done using '$' characters.
Underscores ('_') could be used instead though, since
'_'s can also appear in Statestep identifiers, this
would carry a (very) slight risk of clashes, e.g.,
if there were two variable values called
"open_partially" and "open-partially" in the model
then both would be converted to "_open_partially" in
the code (leading to a compiler error).
-->

<#--
Given the name of a variable in the model (a string),
return the variable identifier in the generated code.
-->
<#function variable_id var_name>
   <#return "$" + var_name/>
</#function>

<#--
Given a value node in a variable definition tree,
i.e., below a 'variable' element, produce the
corresponding unqualified variable value identifier.
-->
<#function value_id def_node>
   <#local id = def_node.@name/>
   <#list def_node?ancestors("value") as group>
      <#local id = group.@name + "$" + id/>
   </#list>
   <#local id = "$" + id/>
   <#if def_node.value?size != 0>
      <#local id = id + "$"/>
   </#if>
   <#return id/>
</#function>

<#--
Given a value node in a variable definition tree,
i.e., below a 'variable' element, produce the
corresponding fully qualified variable value identifier
(i.e., the value id preceded by the variable id).
-->
<#function value_id_full def_node>
   <#local var_name = def_node?ancestors("variable")[0].@name/>
   <#return variable_id(var_name) + "." + value_id(def_node)/>
</#function>

<#--
Return the identifier used for "don't care" sets.
If a variable name is specified then this will be
qualified by the variable identifier in the code,
otherwise it is unqualified.
Parameters:
var_name - name of a variable in the model (a string)
   (optional)
-->
<#function dont_care_id var_name="">
   <#local retval = "DONT_CARE"/>
   <#if var_name != "">
      <#local retval = variable_id(var_name) + "." + retval/>
   </#if>
   <#return retval/>
</#function>

<#--
Given an event node in the model, return the event's
identifier to use in the generated code.
-->
<#function event_id event_node>
   <#return "$" + event_node.@name/>
</#function>

<#--
Given a rule node in the model, return the rule's
identifier to use in the generated code.
-->
<#function rule_id rule_node>
   <#local event_node = rule_node["../.."]/>
   <#return event_node.@name + "$" + rule_node.@name/>
</#function>

<#macro file_prologue>
/* This code has been generated from a Statestep model. */
   <#if (package_name?default("")?trim?length > 0)>

package ${package_name?trim};

   </#if>
</#macro>

<#--
Return the prefix needed to turn an identifier in this
package into a fully qualified name - either a string
ending with a '.' or, if the generated code is not in
any package, an empty string.
-->
<#function package_prefix>
   <#if package_name?default("")?trim?length != 0>
      <#return package_name?trim + "."/>
   <#else>
      <#return ""/>
   </#if>
</#function>

<#--
Variables without at least one value should be ignored.
-->
<#assign usable_variables =
 doc.model.variables["variable[./value]"]
/>

<#--
Given a text node, output a source code prologue comment.
-->
<#macro prologue_comment text indent="   ">
${indent}/**
   <@comment text=text?html indent=indent open_comment=false/>
</#macro>

<#--
Given a text node, output a (non prologue) source code
comment.
-->
<#macro comment text indent="   " open_comment=true>
   <#if open_comment>
${indent}/*
   </#if>
   <@indent_text text=text?replace("*/", "* /") indent=indent/>
${indent}*/
</#macro>

<#--
Given a text node, output the text with indentation
at each line break (works for "\r\n" and "\n" linebreaks).
-->
<#macro indent_text text indent="   ">
${indent}${text?chop_linebreak?replace("\n","\n${indent}")}
</#macro>

<#--
Convert a set of variable values to the corresponding code.
This shouldn't be called for "no change" value sets as
there is no way to express it (instead, "no change" is made
the default in the target row code).
-->
<#macro values_to_code valset>
   <#switch valset.@content>
      <#case "dont_care"
      >${dont_care_id(valset.@variable)}<#break>
      <#case "no_change">
      <#break>
      <#case "list">
         <#local children=valset.value/>
         <#if children?size == 0>
            <#-- an empty list is an implicit "don't care" -->
${dont_care_id(valset.@variable)}<#--
-->
         <#else>
            <#list val_def_nodes(valset) as val
            >${value_id_full(val)}<#if (val_index > 0)>)</#if
            ><#if val_has_next>.or(</#if
            >
            </#list>
         </#if>
      <#break>
   </#switch>
</#macro>

<#--
Given a set of value nodes in a value set,
return the corresponding nodes in the variable
value definition tree.
-->
<#function val_def_nodes value_set>
   <#local result = []/>
   <#local varName = value_set.@variable/>
   <#local var =
    doc.model.variables["variable[@name='${varName}']"][0]
   />
   <#list value_set[".//value[not(./value)]"] as group>
      <#local def_node = var/>
      <#local path_to_group = group?ancestors("value")?reverse
       + [group]/>
      <#list path_to_group as component>
         <#local name=component.@name/>
         <#local def_node = def_node["./value[@name='${name}']"][0]/>
      </#list>
      <#local result = result + [def_node]/>
   </#list>
   <#return result>
</#function>

<#--
This recursive macro converts a parse tree for a
constraint formula or a rule formula into code.
While the generated code will be somewhat inefficient in
some cases (e.g., for a formula like "MyVar -> {=}"), it
must be remembered that the approach used is designed to
ensure more complex formulas like
"A->{a1} && B={b1} || !A->{=} && B={b2|b3}"
are transformed correctly.
This is difficult because formulas use a declarative syntax
which must be mapped to imperative program code.
In particular, for changes to the value of a variable in
a rule formula ("Var -> ..."), some kind of lookahead
or backtracking is needed in case it is later found that
the assignment should not have been made. For this,
the generated code uses recursion.
Simplicity and robustness in the macro are favoured over
elegance or efficiency of the generated code.
-->
<#macro parse_node_to_code node
 fail_cmd="return false;"
 indent="            "
>
   <#if node.@type == "values">
      <#local vals = node.value_set/>
      <#local var_id = variable_id(vals.@variable)/>
      <#if vals.@type == "source">
${indent}if (!comb.getValue(Variable.${var_id}).in(
${indent}   <@values_to_code vals/>))
${indent}{
${indent}   ${fail_cmd}
${indent}}
      <#elseif vals.@type == "target">
         <#if vals.@content == "no_change">
${indent}mask = comb.getValue(Variable.${var_id});
         <#else>
${indent}mask = <@values_to_code vals/>;
         </#if>
${indent}restricted = next.getValues(Variable.${var_id}).and(mask);
${indent}excluded = next.getValues(Variable.${var_id}).andNot(mask);
${indent}if (restricted == null)
${indent}{
${indent}   ${fail_cmd}
${indent}}
${indent}else if (excluded != null)
${indent}{
${indent}   Combination result = apply(comb,
${indent}      new CombinationMask(next).setValues(restricted));
${indent}   if (result != null)
${indent}      return result;
${indent}   return apply(comb,
${indent}      new CombinationMask(next).setValues(excluded));
${indent}}
      <#else>
ERROR: Unexpected value_set type ("${vals.@type}").
      </#if>
   <#elseif node.@type == "not">
      <#local var_id = "inverse_${uniq()}"/>
${indent}boolean ${var_id} = true;
      <@parse_node_to_code node=node.parse_node
       fail_cmd="${var_id} = false;" indent=indent
      />
${indent}if (${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "and">
      <#local var_id = "conjunct_${uniq()}"/>
      <#list node.parse_node as subnode>
         <#if subnode_index == 0>
${indent}boolean ${var_id} = true;
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = false;" indent=indent
            />
         <#else>
${indent}if (${var_id})
${indent}{
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = false;" indent=indent+"   "
            />
${indent}}
         </#if>
      </#list>
${indent}if (!${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "or">
      <#local var_id = "disjunct_${uniq()}"/>
      <#list node.parse_node as subnode>
         <#if subnode_index == 0>
${indent}boolean ${var_id} = true;
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = false;" indent=indent
            />
         <#else>
${indent}if (!${var_id})
${indent}{
${indent}   ${var_id} = true;
${indent}   {
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = false;" indent=indent+"      "
            />
${indent}   }
${indent}}
         </#if>
      </#list>
${indent}if (!${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "implies">
      <#local var_id = "antecedent_${uniq()}"/>
${indent}boolean ${var_id} = true;
      <@parse_node_to_code node=node.parse_node[0]
       fail_cmd="${var_id} = false;" indent=indent
      />
${indent}if (${var_id})
${indent}{
      <@parse_node_to_code node=node.parse_node[1]
       fail_cmd=fail_cmd indent=indent+"   "
      />
${indent}}
   <#else>
ERROR: Unexpected parse_node type ("${node.@type}").
   </#if>
</#macro>

<#--
A function that returns the number of times it has been
called (this number can be appended to identifiers to
ensure uniqueness).
-->
<#function uniq>
   <#assign uniq_count = (uniq_count + 1)?default(1)/>
   <#return uniq_count/>
</#function>

